@@ -165,6 +165,10 @@ val db_open :
   db
 (** [db_open ?mode ?mutex ?cache ?vfs filename] opens the database file
     [filename], and returns a database handle.
+    
+    Special filenames: ":memory:" and "" open an in-memory or temporary
+    database respectively.
+    Behaviour explained here: https://www.sqlite.org/inmemorydb.html
 
     The optional arguments [mode] and [mutex] are only meaningful with SQlite
     versions >= 3.5, [cache] only for versions >= 3.6.18.  For older versions an
@@ -192,7 +196,7 @@ external db_close : db -> bool = "caml_sqlite3_close"
 *)
 
 external enable_load_extension :
-  db -> bool -> bool = "caml_sqlite3_enable_load_extension" "noalloc"
+  db -> bool -> bool = "caml_sqlite3_enable_load_extension"
 (** [enable_load_extension db onoff] enable/disable the sqlite3 load
     extension.  @return [false] if the operation fails, [true]
     otherwise. *)
