@@ -37,6 +37,7 @@
 #include <caml/signals.h>
 
 #include <sqlite3.h>
+#include <pthread.h>
 
 #if __GNUC__ >= 3
 # define inline inline __attribute__ ((always_inline))
@@ -48,7 +49,7 @@
 # define inline
 #endif
 
-#if SQLITE_VERSION_NUMBER >= 3003007
+#if SQLITE_VERSION_NUMBER >= 3003007 && !SQLITE3_DISABLE_LOADABLE_EXTENSIONS
 # define SQLITE_HAS_ENABLE_LOAD_EXTENSION
 #endif
 
@@ -105,6 +106,46 @@ typedef struct stmt_wrap {
 } stmt_wrap;
 
 
+/* Handling of exceptions in user-defined SQL-functions */
+
+/* For propagating exceptions from user-defined SQL-functions */
+static pthread_key_t user_exception_key;
+
+typedef struct user_exception { value exn; } user_exception;
+
+static inline void create_user_exception(value v_exn)
+{
+  user_exception *user_exn = caml_stat_alloc(sizeof(user_exception));
+  user_exn->exn = v_exn;
+  caml_register_global_root(&user_exn->exn);
+  pthread_setspecific(user_exception_key, user_exn);
+}
+
+static inline void destroy_user_exception(void *user_exc_)
+{
+  user_exception *user_exn = user_exc_;
+  caml_remove_global_root(&user_exn->exn);
+  caml_stat_free(user_exn);
+}
+
+static inline void maybe_raise_user_exception(int rc)
+{
+  if (rc == SQLITE_ERROR) {
+    user_exception *user_exn = pthread_getspecific(user_exception_key);
+
+    if (user_exn != NULL) {
+      CAMLparam0();
+      CAMLlocal1(v_exn);
+      v_exn = user_exn->exn;
+      destroy_user_exception(user_exn);
+      pthread_setspecific(user_exception_key, NULL);
+      caml_raise(v_exn);
+      CAMLnoreturn;
+    }
+  }
+}
+
+
 /* Macros to access the wrapper structures stored in the custom blocks */
 
 #define Sqlite3_val(x) (*((db_wrap **) (Data_custom_val(x))))
@@ -120,11 +161,11 @@ static value *caml_sqlite3_RangeError = NULL;
 static inline void raise_with_two_args(value v_tag, value v_arg1, value v_arg2)
 {
   CAMLparam3(v_tag, v_arg1, v_arg2);
-  value v_exc = caml_alloc_small(3, 0);
-  Field(v_exc, 0) = v_tag;
-  Field(v_exc, 1) = v_arg1;
-  Field(v_exc, 2) = v_arg2;
-  caml_raise(v_exc);
+  value v_exn = caml_alloc_small(3, 0);
+  Field(v_exn, 0) = v_tag;
+  Field(v_exn, 1) = v_arg1;
+  Field(v_exn, 2) = v_arg2;
+  caml_raise(v_exn);
   CAMLnoreturn;
 }
 
@@ -215,6 +256,7 @@ CAMLprim value caml_sqlite3_init(value __unused v_unit)
   caml_sqlite3_InternalError = caml_named_value("Sqlite3.InternalError");
   caml_sqlite3_Error = caml_named_value("Sqlite3.Error");
   caml_sqlite3_RangeError = caml_named_value("Sqlite3.RangeError");
+  pthread_key_create(&user_exception_key, destroy_user_exception);
   return Val_unit;
 }
 
@@ -285,10 +327,10 @@ static inline value copy_not_null_string_array(const char** strs, int len)
   }
 }
 
-static inline value safe_copy_string_array(const char** strs, int len)
+static inline value safe_copy_header_strings(const char** strs, int len)
 {
   value v_res = copy_not_null_string_array(strs, len);
-  if (v_res == (value) NULL) raise_sqlite3_Error("Null element in row");
+  if (v_res == (value) NULL) raise_sqlite3_Error("Null element in header");
   return v_res;
 }
 
@@ -298,14 +340,15 @@ static inline value safe_copy_string_array(const char** strs, int len)
 static inline void ref_count_finalize_dbw(db_wrap *dbw)
 {
   if (--dbw->ref_count == 0) {
-    user_function *link;
-    for (link = dbw->user_functions; link != NULL; link = link->next) {
+    user_function *link, *next;
+    for (link = dbw->user_functions; link != NULL; link = next) {
       caml_remove_generational_global_root(&link->v_fun);
-      free(link);
+      next = link->next;
+      caml_stat_free(link);
     }
     dbw->user_functions = NULL;
     sqlite3_close(dbw->db);
-    free(dbw);
+    caml_stat_free(dbw);
   }
 }
 
@@ -377,11 +420,11 @@ CAMLprim value caml_sqlite3_open(
   caml_enter_blocking_section();
 #ifdef SQLITE_HAS_OPEN_V2
     res = sqlite3_open_v2(file, &db, flags, vfs);
-    free(vfs);
+    caml_stat_free(vfs);
 #else
     res = sqlite3_open(file, &db);
 #endif
-    free(file);
+    caml_stat_free(file);
   caml_leave_blocking_section();
 
   if (res) {
@@ -477,7 +520,7 @@ static inline int exec_callback(
     v_row = copy_string_option_array((const char **) row, num_columns);
 
     Begin_roots1(v_row);
-      v_header = safe_copy_string_array((const char **) header, num_columns);
+      v_header = safe_copy_header_strings((const char **) header, num_columns);
     End_roots();
 
     v_ret = caml_callback2_exn(*cbx->cbp, v_row, v_header);
@@ -517,10 +560,11 @@ CAMLprim value caml_sqlite3_exec(value v_db, value v_maybe_cb, value v_sql)
 
   caml_enter_blocking_section();
     rc = sqlite3_exec(dbw->db, sql, cb, (void *) &cbx, NULL);
-    free(sql);
+    caml_stat_free(sql);
   caml_leave_blocking_section();
 
   if (rc == SQLITE_ABORT) caml_raise(*cbx.exn);
+  maybe_raise_user_exception(rc);
 
   CAMLreturn(Val_rc(rc));
 }
@@ -566,10 +610,11 @@ CAMLprim value caml_sqlite3_exec_no_headers(value v_db, value v_cb, value v_sql)
   caml_enter_blocking_section();
     rc =
       sqlite3_exec(dbw->db, sql, exec_callback_no_headers, (void *) &cbx, NULL);
-    free(sql);
+    caml_stat_free(sql);
   caml_leave_blocking_section();
 
   if (rc == SQLITE_ABORT) caml_raise(*cbx.exn);
+  maybe_raise_user_exception(rc);
 
   CAMLreturn(Val_rc(rc));
 }
@@ -584,10 +629,14 @@ static inline int exec_not_null_callback(
 
     v_row = copy_not_null_string_array((const char **) row, num_columns);
 
-    if (v_row == (value) NULL) return 1;
+    if (v_row == (value) NULL)
+    {
+      caml_enter_blocking_section();
+      return 1;
+    }
 
     Begin_roots1(v_row);
-      v_header = safe_copy_string_array((const char **) header, num_columns);
+      v_header = safe_copy_header_strings((const char **) header, num_columns);
     End_roots();
 
     v_ret = caml_callback2_exn(*cbx->cbp, v_row, v_header);
@@ -622,13 +671,15 @@ CAMLprim value caml_sqlite3_exec_not_null(value v_db, value v_cb, value v_sql)
   caml_enter_blocking_section();
     rc =
       sqlite3_exec(dbw->db, sql, exec_not_null_callback, (void *) &cbx, NULL);
-    free(sql);
+    caml_stat_free(sql);
   caml_leave_blocking_section();
 
   if (rc == SQLITE_ABORT) {
     if (*cbx.exn != 0) caml_raise(*cbx.exn);
     else raise_sqlite3_Error("Null element in row");
   }
+  maybe_raise_user_exception(rc);
+
   CAMLreturn(Val_rc(rc));
 }
 
@@ -641,7 +692,12 @@ static inline int exec_not_null_no_headers_callback(
   caml_leave_blocking_section();
 
     v_row = copy_not_null_string_array((const char **) row, num_columns);
-    if (v_row == (value) NULL) return 1;
+    if (v_row == (value) NULL)
+    {
+      caml_enter_blocking_section();
+      return 1;
+    }
+
     v_ret = caml_callback_exn(*cbx->cbp, v_row);
 
     if (Is_exception_result(v_ret)) {
@@ -676,13 +732,15 @@ CAMLprim value caml_sqlite3_exec_not_null_no_headers(
     rc =
       sqlite3_exec(
         dbw->db, sql, exec_not_null_no_headers_callback, (void *) &cbx, NULL);
-    free(sql);
+    caml_stat_free(sql);
   caml_leave_blocking_section();
 
   if (rc == SQLITE_ABORT) {
     if (*cbx.exn != 0) caml_raise(*cbx.exn);
     else raise_sqlite3_Error("Null element in row");
   }
+  maybe_raise_user_exception(rc);
+
   CAMLreturn(Val_rc(rc));
 }
 
@@ -694,9 +752,9 @@ static inline void finalize_stmt_gc(value v_stmt)
   stmt_wrap *stmtw = Sqlite3_stmtw_val(v_stmt);
   sqlite3_stmt *stmt = stmtw->stmt;
   if (stmt) sqlite3_finalize(stmt);
-  if (stmtw->sql) free(stmtw->sql);
+  if (stmtw->sql) caml_stat_free(stmtw->sql);
   ref_count_finalize_dbw(stmtw->db_wrap);
-  free(stmtw);
+  caml_stat_free(stmtw);
 }
 
 CAMLprim value caml_sqlite3_stmt_finalize(value v_stmt)
@@ -992,14 +1050,16 @@ static inline value caml_sqlite3_wrap_values(int argc, sqlite3_value **args)
   }
 }
 
-static inline void exception_result(sqlite3_context *ctx)
+static inline void exception_result(sqlite3_context *ctx, value v_res)
 {
+  value v_exn = Extract_exception(v_res);
+  create_user_exception(v_exn);
   sqlite3_result_error(ctx, "OCaml callback raised an exception", -1);
 }
 
 static inline void set_sqlite3_result(sqlite3_context *ctx, value v_res)
 {
-  if (Is_exception_result(v_res)) exception_result(ctx);
+  if (Is_exception_result(v_res)) exception_result(ctx, v_res);
   else if (Is_long(v_res)) sqlite3_result_null(ctx);
   else {
     value v = Field(v_res, 0);
@@ -1051,7 +1111,7 @@ static inline void caml_sqlite3_user_function_step(
     }
     v_args = caml_sqlite3_wrap_values(argc, argv);
     v_res = caml_callback2_exn(Field(data->v_fun, 2), agg_ctx->v_acc, v_args);
-    if (Is_exception_result(v_res)) exception_result(ctx);
+    if (Is_exception_result(v_res)) exception_result(ctx, v_res);
     else agg_ctx->v_acc = v_res;
   caml_enter_blocking_section();
 }
@@ -1078,7 +1138,7 @@ static inline void unregister_user_function(db_wrap *db_data, value v_name)
       if (prev == NULL) db_data->user_functions = link->next;
       else prev->next = link->next;
       caml_remove_generational_global_root(&link->v_fun);
-      free(link);
+      caml_stat_free(link);
       break;
     }
     prev = link;
@@ -1198,4 +1258,3 @@ CAMLprim value caml_sqlite3_changes(value v_db)
   check_db(dbw, "changes");
   return Val_int(sqlite3_changes(dbw->db));
 }
-
